@@ -1,5 +1,5 @@
 // Copyright (c) 2020-2023 Christopher Davis
-// Copyright (c) 2022-2024 Sophie Herold
+// Copyright (c) 2022-2025 Sophie Herold
 // Copyright (c) 2022 Maximiliano Sandoval R
 // Copyright (c) 2023 FineFindus
 // Copyright (c) 2023 Huan Nguyen
@@ -37,15 +37,15 @@ use adw::subclass::prelude::*;
 use anyhow::Context;
 use ashpd::desktop::{wallpaper, ResponseError};
 use ashpd::{Error, WindowIdentifier};
-use glib::translate::IntoGlib;
 use glib::{clone, Properties};
 use gtk::CompositeTemplate;
 
 use crate::decoder::DecoderError;
 use crate::deps::*;
-use crate::file_model::LpFileModel;
+use crate::file_model::{FileEvent, LpFileModel};
 use crate::util::gettext::*;
-use crate::util::{Direction, Position};
+use crate::util::root::ParentWindow;
+use crate::util::{self, Direction, Position};
 use crate::widgets::{LpImage, LpImagePage, LpPrint, LpSlidingView};
 
 // The number of pages we want to buffer
@@ -63,18 +63,38 @@ mod imp {
         /// Direct child of this Adw::Bin
         #[template_child]
         pub(super) bin_child: TemplateChild<gtk::Widget>,
+
         /// overlayed controls
         #[template_child]
         pub(super) controls_box_start: TemplateChild<gtk::Widget>,
         #[template_child]
+        pub(super) zoom_toggle: TemplateChild<gtk::ToggleButton>,
+        #[template_child]
         pub(super) controls_box_start_events: TemplateChild<gtk::EventControllerMotion>,
+
         /// overlayed controls
         #[template_child]
         pub(super) controls_box_end: TemplateChild<gtk::Widget>,
         #[template_child]
         pub(super) controls_box_end_events: TemplateChild<gtk::EventControllerMotion>,
         #[template_child]
-        pub(super) fullscreen_button: TemplateChild<gtk::Button>,
+        pub(super) zoom_menu_button: TemplateChild<gtk::MenuButton>,
+        #[template_child]
+        pub(super) zoom_to_list: TemplateChild<gtk::ListBox>,
+        #[template_child]
+        pub(super) zoom_to_300: TemplateChild<gtk::ListBoxRow>,
+        #[template_child]
+        pub(super) zoom_to_200: TemplateChild<gtk::ListBoxRow>,
+        #[template_child]
+        pub(super) zoom_to_100: TemplateChild<gtk::ListBoxRow>,
+        #[template_child]
+        pub(super) zoom_to_66: TemplateChild<gtk::ListBoxRow>,
+        #[template_child]
+        pub(super) zoom_to_50: TemplateChild<gtk::ListBoxRow>,
+        #[template_child]
+        pub(super) zoom_value: TemplateChild<gtk::Entry>,
+        #[template_child]
+        pub(super) submit_zoom: TemplateChild<gtk::Button>,
 
         #[template_child]
         pub sliding_view: TemplateChild<LpSlidingView>,
@@ -98,6 +118,12 @@ mod imp {
 
         #[property(get = Self::is_next_available)]
         _is_next_available: PhantomData<bool>,
+
+        #[property(get, set)]
+        zoom_toggle_state: Cell<bool>,
+
+        #[property(get, set, nullable)]
+        pub delayed_current_file: RefCell<Option<gio::File>>,
     }
 
     #[glib::object_subclass]
@@ -148,32 +174,113 @@ mod imp {
                 move || obj.target_page_reached()
             ));
 
-            let signal_group = obj.current_image_signals();
+            let current_image_signal_group = obj.current_image_signals();
 
             let view = &*obj;
-            signal_group.connect_closure(
-                "notify::file",
+
+            current_image_signal_group.connect_closure(
+                "notify::zoom-target",
                 true,
                 glib::closure_local!(
                     #[watch]
                     view,
-                    move |_: LpImage, _: glib::ParamSpec| {
-                        view.current_image_path_changed();
+                    move |img: LpImage, _: glib::ParamSpec| {
+                        let mut percent =
+                            ((img.zoom_target() * 10_000.).round() / 100.).to_string();
+
+                        if let Some(decimal) = util::locale_settings().decimal_point {
+                            percent = percent.replace('.', &decimal);
+                        }
+
+                        view.imp()
+                            .zoom_value
+                            .set_text(&gettext_f("{} %", [percent]));
+                        view.imp().zoom_value.select_region(0, -1);
                     }
                 ),
             );
 
-            signal_group.connect_closure(
-                "notify::is-deleted",
-                true,
-                glib::closure_local!(
-                    #[watch]
-                    view,
-                    move |_: LpImage, _: glib::ParamSpec| {
-                        view.current_image_path_changed();
+            self.zoom_value.connect_changed(glib::clone!(
+                #[weak]
+                obj,
+                move |_| {
+                    if let Some(image) = obj.current_image() {
+                        let imp = obj.imp();
+
+                        // Only make submit button sensitive if entry value is tifferent from
+                        // current zoom value
+                        let sensitive = if let Some(value) = obj.zoom_value_f64() {
+                            let a = (value * 10_000.).round() / 10_000.;
+                            let b = (image.zoom_target() * 10_000.).round() / 10_000.;
+                            a != b
+                        } else {
+                            false
+                        };
+
+                        imp.submit_zoom.set_sensitive(sensitive);
+                        if sensitive {
+                            imp.submit_zoom.add_css_class("suggested-action");
+                        } else {
+                            imp.submit_zoom.remove_css_class("suggested-action");
+                        }
                     }
-                ),
-            );
+                }
+            ));
+
+            self.zoom_toggle.connect_clicked(clone!(
+                #[weak]
+                obj,
+                move |button| {
+                    if let Some(page) = obj.current_page() {
+                        if button.is_active() {
+                            page.image().zoom_in_center()
+                        } else {
+                            page.image().zoom_best_fit()
+                        };
+                    }
+                }
+            ));
+
+            self.zoom_to_list.connect_row_activated(clone!(
+                #[weak]
+                obj,
+                move |_, row| {
+                    let imp = obj.imp();
+                    let value = if row == &*imp.zoom_to_300 {
+                        3.
+                    } else if row == &*imp.zoom_to_200 {
+                        2.
+                    } else if row == &*imp.zoom_to_100 {
+                        1.
+                    } else if row == &*imp.zoom_to_66 {
+                        0.66
+                    } else if row == &*imp.zoom_to_50 {
+                        0.5
+                    } else {
+                        log::error!("Activated unknown entry in zoom to popover.");
+                        return;
+                    };
+
+                    if let Some(image) = obj.current_image() {
+                        image.zoom_to_center_no_best_fit(value);
+                    }
+
+                    // Close popover after click
+                    imp.zoom_menu_button.set_active(false);
+                }
+            ));
+
+            self.submit_zoom.connect_clicked(clone!(
+                #[weak]
+                obj,
+                move |_| {
+                    if let Some(page) = obj.current_page() {
+                        if let Some(value) = obj.zoom_value_f64() {
+                            page.image().zoom_to_no_best_fit(value);
+                        }
+                    }
+                }
+            ));
 
             self.drag_source.set_exclusive(true);
 
@@ -312,16 +419,22 @@ impl LpImageView {
     }
 
     fn load_files(&self, files: Vec<gio::File>) {
-        let sliding_view = self.sliding_view().editor();
-
         if let Some(first) = files.first().cloned() {
-            let model = LpFileModel::from_files(files);
-            self.set_model(model);
-            let page = self.new_image_page(&first);
+            glib::spawn_future_local(glib::clone!(
+                #[strong(rename_to=obj)]
+                self,
+                async move {
+                    let sliding_view = obj.sliding_view().editor();
 
-            sliding_view.clear_lazy();
-            sliding_view.append(&page);
-            self.update_sliding_view(&first);
+                    let model = LpFileModel::from_files(files).await;
+                    obj.set_model(model);
+                    let page = obj.new_image_page(&first);
+
+                    sliding_view.clear_lazy();
+                    sliding_view.append(&page);
+                    obj.update_sliding_view(&first);
+                }
+            ));
         } else {
             log::error!("File list was empty");
         }
@@ -347,38 +460,31 @@ impl LpImageView {
             return;
         }
 
-        let model = LpFileModel::from_file(file.clone());
-        self.set_model(model);
-        log::debug!("New model created");
+        glib::spawn_future_local(glib::clone!(
+            #[strong(rename_to = obj)]
+            self,
+            async move {
+                let model = LpFileModel::from_file(file.clone()).await;
+                obj.set_model(model);
+                log::debug!("New model created");
 
-        self.update_sliding_view(&file);
+                obj.update_sliding_view(&file);
 
-        // List other files in directory
-        if let Some(directory) = directory {
-            glib::spawn_future_local(glib::clone!(
-                #[weak(rename_to = obj)]
-                self,
-                async move {
+                // List other files in directory
+                if let Some(directory) = directory {
                     if let Err(err) = obj.model().load_directory(directory.clone()).await {
                         log::warn!("Failed to load directory: {}", err.root_cause());
-                        obj.activate_action(
-                            "win.show-toast",
-                            Some(
-                                &(
-                                    format!("{} {}", err, err.root_cause()),
-                                    adw::ToastPriority::High.into_glib(),
-                                )
-                                    .to_variant(),
-                            ),
-                        )
-                        .unwrap();
+                        obj.window_show_toast(
+                            &format!("{} {}", err, err.root_cause()),
+                            adw::ToastPriority::High,
+                        );
                         return;
                     }
 
                     obj.update_sliding_view(&file);
                 }
-            ));
-        }
+            }
+        ));
     }
 
     /// Move forward or backwards
@@ -419,7 +525,11 @@ impl LpImageView {
         }
 
         let Some(new_index) = self.model().index_of(new_file) else {
-            log::warn!("Could not navigate to file {}", new_file.uri());
+            log::debug!(
+                "Could not navigate to file that's not in model '{}': Delaying insertion",
+                new_file.uri()
+            );
+            self.set_delayed_current_file(Some(new_file));
             return;
         };
 
@@ -445,6 +555,8 @@ impl LpImageView {
         log::debug!("Scrolling to page for {}", new_file.uri());
         self.imp().preserve_content.set(true);
         sliding_view.animate_to(&page);
+
+        self.set_delayed_current_file(gio::File::NONE);
     }
 
     /// Ensures the sliding view contains the correct images
@@ -494,6 +606,8 @@ impl LpImageView {
         } else {
             self.sliding_view().instant_to(&current_page);
         }
+
+        self.set_delayed_current_file(gio::File::NONE);
     }
 
     fn sliding_view(&self) -> LpSlidingView {
@@ -554,16 +668,16 @@ impl LpImageView {
         model.connect_changed(glib::clone!(
             #[weak(rename_to = obj)]
             self,
-            move || obj.model_content_changed_cb()
+            move |change| obj.model_content_changed_cb(change)
         ));
         self.imp().model.replace(model);
     }
 
     /// Handle files are added or removed from directory
-    fn model_content_changed_cb(&self) {
+    fn model_content_changed_cb(&self, file_event: &FileEvent) {
         // Animate to image that was restored from trash
         if let Some(trash_restore) = self.trash_restore() {
-            if self.model().contains(&trash_restore) {
+            if self.model().contains_file(&trash_restore) {
                 self.set_trash_restore(gio::File::NONE);
                 self.update_sliding_view(&trash_restore);
                 self.scroll_sliding_view(&trash_restore, true);
@@ -575,30 +689,38 @@ impl LpImageView {
             return;
         };
 
-        // LpImage did not get the update yet
-        // Update will be handled by current_image_path_changed
-        if !self.model().contains(&current_file) {
-            return;
-        }
-
-        self.update_sliding_view(&current_file);
-    }
-
-    /// Handle current image being moved or deleted
-    ///
-    /// This is handled separately since we want to delete animations and
-    /// want to still show the same image if renamed.
-    fn current_image_path_changed(&self) {
-        if let Some(image) = self.current_page().map(|x| x.image()) {
-            if image.is_deleted() {
+        match file_event {
+            FileEvent::Removed(removed) if removed == &current_file.uri() => {
                 self.sliding_view().scroll_to_neighbor();
+                return;
             }
+            FileEvent::Moved(src, target) => {
+                let target_file = gio::File::for_uri(target);
+                // Set new filename for image
+                if let Some(page) = self.sliding_view().get(&gio::File::for_uri(src)) {
+                    page.image().init(&target_file);
+                }
+                // File that delayed move is set for is now available
+                if let Some(delayed_current_file) = self.delayed_current_file() {
+                    if target_file.equal(&delayed_current_file) {
+                        self.navigate_to_file(&delayed_current_file);
+                    }
+                }
+            }
+            FileEvent::New(new) => {
+                // File that delayed move is set for is now available
+                if let Some(delayed_current_file) = self.delayed_current_file() {
+                    let new_file = gio::File::for_uri(new);
+                    if new_file.equal(&delayed_current_file) {
+                        self.navigate_to_file(&delayed_current_file);
+                    }
+                }
+            }
+            _ => {}
         }
 
-        if let Some(current_file) = self.current_file() {
-            if self.model().contains(&current_file) && !self.imp().preserve_content.get() {
-                self.update_sliding_view(&current_file);
-            }
+        if !self.imp().preserve_content.get() {
+            self.update_sliding_view(&current_file);
         }
     }
 
@@ -624,14 +746,11 @@ impl LpImageView {
         self.imp()
             .sliding_view
             .current_page()
-            .map(|x| x.file().uri())
+            .map(|x: LpImagePage| x.file().uri())
     }
 
     pub fn current_file(&self) -> Option<gio::File> {
-        self.imp()
-            .sliding_view
-            .current_page()
-            .and_then(|x| x.image().file())
+        self.current_uri().map(|x| gio::File::for_uri(&x))
     }
 
     pub fn drag_source(&self) -> gtk::DragSource {
@@ -678,16 +797,6 @@ impl LpImageView {
         }
     }
 
-    pub fn on_fullscreen_changed(&self, is_fullscreened: bool) {
-        let icon = if is_fullscreened {
-            "view-restore-symbolic"
-        } else {
-            "view-fullscreen-symbolic"
-        };
-
-        self.imp().fullscreen_button.set_icon_name(icon);
-    }
-
     pub fn rotate_image(&self, angle: f64) {
         if let Some(current_page) = self.current_page() {
             current_page.image().rotate_by(angle);
@@ -696,8 +805,8 @@ impl LpImageView {
 
     pub async fn set_background(&self) -> anyhow::Result<()> {
         let uri = self
-            .current_file()
-            .and_then(|f| url::Url::parse(f.uri().as_str()).ok())
+            .current_uri()
+            .and_then(|x| url::Url::parse(x.as_str()).ok())
             .context("Invalid URL for background image")?;
         let native = self.native().context("View should have a GtkNative")?;
         let id = WindowIdentifier::from_native(&native).await;
@@ -710,33 +819,18 @@ impl LpImageView {
             .await
             .and_then(|req| req.response())
         {
-            Ok(_) => self
-                .activate_action(
-                    "win.show-toast",
-                    Some(
-                        &(
-                            // Translators: This is a toast notification, informing the user that
-                            // an image has been set as background.
-                            gettext("Set as background."),
-                            adw::ToastPriority::High.into_glib(),
-                        )
-                            .to_variant(),
-                    ),
-                )
-                .unwrap(),
+            Ok(_) => self.window_show_toast(
+                // Translators: This is a toast notification, informing the user that
+                // an image has been set as background.
+                &gettext("Set as background."),
+                adw::ToastPriority::High,
+            ),
             Err(err) => {
                 if !matches!(err, Error::Response(ResponseError::Cancelled)) {
-                    self.activate_action(
-                        "win.show-toast",
-                        Some(
-                            &(
-                                gettext("Could not set background."),
-                                adw::ToastPriority::High.into_glib(),
-                            )
-                                .to_variant(),
-                        ),
-                    )
-                    .unwrap();
+                    self.window_show_toast(
+                        &gettext("Could not set background."),
+                        adw::ToastPriority::High,
+                    );
                 }
             }
         };
@@ -770,4 +864,30 @@ impl LpImageView {
 
         Ok(())
     }
+
+    pub fn zoom_menu_button(&self) -> gtk::MenuButton {
+        self.imp().zoom_menu_button.clone()
+    }
+
+    pub fn zoom_toggle(&self) -> gtk::ToggleButton {
+        self.imp().zoom_toggle.clone()
+    }
+
+    pub fn zoom_value_f64(&self) -> Option<f64> {
+        let mut value = self.imp().zoom_value.text().to_string();
+        let locale_settings = util::locale_settings();
+
+        if let Some(thousands) = locale_settings.thousands_sep {
+            value = value.replace(&thousands, "");
+        }
+
+        if let Some(decimal) = locale_settings.decimal_point {
+            value = value.replace(&decimal, ".");
+        }
+
+        value = value.replace('%', "");
+        value = value.trim().to_string();
+
+        value.parse::<f64>().ok().map(|x| x / 100.)
+    }
 }
